syntax = "proto3";
package namespace.general;

message Error {
    string title = 1;
    string message = 2;
    string stack = 3;
}

message ErrorList {
    repeated namespace.general.Error elements = 1;
}